# Generated by Django 4.2.9 on 2024-01-27 07:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sequence', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sequence',
            name='name',
            field=models.CharField(max_length=255, verbose_name='Name'),
        ),
    ]



import logging
from datetime import datetime, date, timedelta
from django.utils.translation import gettext_lazy as _
from django import template
from django.db.models import Manager
from django.forms import Form, ModelForm
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from accrete.contrib.ui import TableField, TableFieldType

_logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(name='combine_templates')
def combine_templates(template_name, request=None):
    html = ''
    for app in settings.INSTALLED_APPS:
        try:
            html += render_to_string(
                f'{app.split(".")[-1]}/{template_name}', request=request
            )
        except template.TemplateDoesNotExist:
            continue
    return mark_safe(html)


@register.filter(name='field_attr')
def get_attr_from_field(obj: object, field: TableField):

    def return_save(attr):
        if isinstance(attr, Manager):
            return attr.related_model
        if attr is None:
            return None
        if isinstance(attr, str):
            return attr
        return attr
    f_name = field.name
    if field.field_type == TableFieldType.CHOICE_DISPLAY:
        f_name = f'get_{f_name}_display'

    try:
        attribute = getattr(obj, f_name, False)
    except (AttributeError,):
        _logger.warning(f'Object {obj} has no attribute {f_name}')
        attribute = None
    return return_save(attribute)


@register.filter(name='get_attr')
def get_attr_from_string(param: object, value: str):

    def return_save(attr):
        if isinstance(attr, Manager):
            return attr.related_model
        if attr is None:
            return None
        if isinstance(attr, str):
            return attr
        return attr

    try:
        attribute = getattr(param, value, False)
    except (AttributeError,):
        _logger.warning(f'Object {param} has no attribute {value}')
        attribute = None
    return return_save(attribute)


@register.filter(name='related_obj_url')
def related_obj_url(param: object, value: str):
    attr = getattr(param, value, False)
    if attr and hasattr(attr, 'get_absolute_url'):
        return attr.get_absolute_url()
    return '#'


@register.filter(name='message_class')
def message_class(param):
    if param.level == 25:
        return 'is-success'
    if param.level == 30:
        return 'is-warning'
    if param.level == 40:
        return 'is-danger'


@register.filter(name='non_field_error_code')
def non_field_error_code(form: [Form|ModelForm], code: str) -> bool:
    if not form:
        return False
    for error in form.non_field_errors().as_data():
        if error.code and error.code == code:
            return True
    return False


@register.filter(name='timedelta_cast')
def timedelta_cast(td: timedelta, code: str) -> str | None:
    if not isinstance(td, timedelta):
        return None
    codes = ['days', 'hours', 'minutes', 'seconds', 'microseconds']
    if code not in codes:
        return None
    return str(getattr(td, code))


@register.filter(name='weekday')
def datetime_to_weekday(dt: datetime|date, default=None):
    if dt is None:
        return default
    mapping = {
        1: _('Mon'),
        2: _('Tue'),
        3: _('Wed'),
        4: _('Thu'),
        5: _('Fri'),
        6: _('Sat'),
        7: _('Sun')
    }
    return mapping[dt.isoweekday()]



from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth import views, update_session_auth_hash
from django.contrib.auth.decorators import login_required
from django.contrib import messages
from django.shortcuts import redirect, render, reverse, resolve_url
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from accrete.utils import save_form
from accrete.contrib import ui
from .forms import UserForm, ChangePasswordForm, ChangeEmailForm


class LoginView(views.LoginView):

    form_class = AuthenticationForm
    template_name = 'user/login.html'
    redirect_authenticated_user = True

    def form_invalid(self, form):
        user = form.get_user()
        if user is not None and not user.is_active:
            ctx = {'to_confirm': True}
            if self.extra_context:
                self.extra_context.update(ctx)
            else:
                self.extra_context = ctx
        return super().form_invalid(form)


class LogoutView(views.LogoutView):

    def get_success_url(self):
        return resolve_url(settings.LOGIN_URL)


@login_required()
def user_detail(request):
    ctx = ui.DetailContext(
        title=_('Preferences'),
        object=request.user,
        breadcrumbs=[],
        actions=[
            ui.ClientAction(_('Edit'), url=reverse('user:edit')),
            ui.ClientAction(_('Change E-Mail'), url=reverse('user:edit_email')),
            ui.ClientAction(_('Change Password'), url=reverse('user:edit_password'))
        ]
    ).dict()
    return render(request, 'user/user_detail.html', ctx)


@login_required()
def user_edit(request):
    form = UserForm(
        initial={'language_code': request.user.language_code},
        instance=request.user
    )
    if request.method == 'POST':
        form = save_form(UserForm(request.POST, instance=request.user))
        if form.is_saved:
            return redirect('user:detail')
    ctx = ui.FormContext(
        title=_('Preferences'),
        form=form,
        form_id='form',
        actions=ui.form_actions(reverse('user:detail'))
    ).dict()
    return render(request, 'user/user_form.html', ctx)


@login_required()
def user_change_password(request):
    form = ChangePasswordForm(instance=request.user)
    if request.method == 'POST':
        form = save_form(ChangePasswordForm(request.POST, instance=request.user))
        if form.is_saved:
            update_session_auth_hash(request, form.instance)
            messages.add_message(
                request, messages.SUCCESS,
                str(_('Your password has been changed.')),
                fail_silently=True
            )
            return redirect('user:detail')
    ctx = ui.FormContext(
        title=_('Change Password'),
        form=form,
        form_id='form',
        actions=ui.form_actions(reverse('user:detail'))
    ).dict()
    return render(request, 'user/change_password.html', ctx)


@login_required()
def user_change_email(request):
    form = ChangeEmailForm(instance=request.user)
    if request.method == 'POST':
        form = save_form(ChangeEmailForm(request.POST, instance=request.user))
        if form.is_saved:
            messages.add_message(
                request, messages.SUCCESS,
                str(_('Your email address has been changed.')),
                fail_silently=True
            )
            return redirect('user:detail')

    ctx = ui.FormContext(
        title=_('Change Email'),
        form=form,
        form_id='form',
        actions=ui.form_actions(reverse('user:detail'))
    ).dict()
    return render(request, 'user/change_email.html', ctx)



from django.db import models
from django.utils.translation import gettext_lazy as _
from accrete.models import TenantModel
from accrete.managers import TenantManager
from accrete.tenant import get_tenant


class TagManager(TenantManager):

    def get_queryset(self):
        queryset = super().get_queryset()
        if group := not getattr(self.model, 'tag_group', None):
            print(group)
        if group := getattr(self.model, 'tag_group', False) and get_tenant():
            return queryset.filter(tag_group=group)
        return queryset


class Tag(TenantModel):

    class Meta:
        verbose_name = _('Tag')
        verbose_name_plural = _('Tags')
        db_table = 'accrete_tag'
        default_related_name = 'tags'
        constraints= [
            models.UniqueConstraint(
                fields=['tenant', 'name'], name='unique_tag_name_per_tenant'
            ),
            models.UniqueConstraint(
                fields=['tenant', 'tag_group'], name='unique_tag_group_per_tenant'
            )
        ]

    objects = TagManager()

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=50
    )

    color = models.CharField(
        verbose_name=_('Color'),
        max_length=7
    )

    tag_group = models.CharField(
        verbose_name=_('Group'),
        max_length=255
    )



import re
import logging
from uuid import uuid4
from typing import Type
from django.db import transaction
from django.forms import BaseFormSet, Form, ModelForm

_logger = logging.getLogger(__name__)


def save_form(form: [Form|ModelForm], commit=True, reraise=False) -> [Form|ModelForm]:
    form.is_saved = False
    form.save_error = None
    form.save_error_id = None
    try:
        if form.is_valid():
            with transaction.atomic():
                form.save(commit=commit)
            form.is_saved = True
    except Exception as e:
        form.save_error = repr(e)
        error_id = str(uuid4())[:8]
        _logger.exception(f'{error_id}: {e}')
        form.save_error_id = error_id
        if reraise:
            raise e
    return form


def save_forms(form, inline_formsets: list = None, commit=True, reraise: bool = False):

    def handle_error(error):
        form.save_error = repr(error)
        error_id = str(uuid4())[:8]
        _logger.exception(f'{error_id}: {error}')
        form.save_error_id = error_id

    form.is_saved = False
    form.save_error = None
    form.save_error_id = None
    form.inline_forms = inline_formsets

    try:
        form.is_valid()
        inlines_valid = all([
            inline_formset.is_valid() for inline_formset in inline_formsets
        ])
    except Exception as e:
        handle_error(e)
        if reraise:
            raise e
        return form

    if not form.is_valid() or not inlines_valid:
        return form

    try:
        with transaction.atomic():
            form.save(commit=commit)
            for inline_formset in inline_formsets:
                inline_formset.save(commit=commit)
    except Exception as e:
        handle_error(e)
        if reraise:
            raise e
        return form

    form.is_saved = True
    return form


def inline_vals_from_post(post: dict, prefix: str) -> list[dict]:
    post_keys = set(re.findall(f'{prefix}-[0-9]+', ', '.join(post.keys())))
    initial_data = {
        post_key: {}
        for post_key in post_keys if not post.get(f'{post_key}-DELETE')
    }
    for key, val in post.items():
        post_key = '-'.join(key.split('-')[:-1])
        if post_key not in initial_data:
            continue
        field_name = key.split('-')[-1]
        initial_data[post_key].update({field_name: val})
    return [val for val in initial_data.values()]


def extend_formset(formset_class, post: dict, data: list[dict]|dict, **formset_kwargs) -> Type[BaseFormSet]:
    formset = formset_class(post, **formset_kwargs)
    if not formset.is_valid():
        return formset
    form_data = post.copy()
    if isinstance(data, dict):
        data = [data]
    prefix = formset_kwargs.get('prefix', 'form')
    total = int(form_data[f'{prefix}-TOTAL_FORMS']) - 1
    for item in data:
        total += 1
        form_data.update({f'{prefix}-{total}-{key}': value for key, value in item.items()})
    form_data[f'{prefix}-TOTAL_FORMS'] = total + 1
    formset = formset_class(form_data, **formset_kwargs)
    for form in formset:
        form._errors = {}
    return formset
